/*
 *	Copyright (c) 1999-2002 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <paths.h>

#include "threads.h"
#include "netmon.h"
#include "mib.h"

char copyright[] = "Copyright (c) 1999-2004 Rinet Corp., Novosibirsk, Russia\n"
		   "Copyright (c) 2011-2013 Andrey V. Elsukov <bu7cher@yandex.ru>";

/*
 * Forward declarations.
 */

void mainloop __P((void));
void vers __P((void));
void usage __P((void));

static void cleanup_handler __P((int));
static void watchdog_handler __P((int));

/*
 * Global varibales.
 */
char *program_name;		/* my self name */
char *config_file;		/* netmond config file */
CONFIG *cf;			/* current configuration */
int Cflag;			/* command line switches */
pid_t mypid;			/* my self PID */
int syslog_initialized;		/* syslog ready to use */
int syslog_facility;		/* current syslog facility */
int compatibility_flag = 1;	/* version backward compatibility flag.
                                   default - work like previous version  */

#ifdef	HAVE_PTHREAD
pthread_t main_thr;
int busy_threads;
pthread_attr_t detached_attr;
pthread_attr_t joinable_attr;
pthread_mutex_t gethostby_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t localtime_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t variables_lock = PTHREAD_MUTEX_INITIALIZER;
#else
static char **Argv;		/* pointer to argument vector */
static char *LastArgv;		/* end of argv */
#endif

static int watchdog_timeout;
static int watchdog_pending;
static int siglist[] = {
	SIGCHLD, SIGALRM, SIGHUP, SIGINT, SIGPIPE, SIGTERM, 0
};
#define	RECONFIG_SOFT	1
#define	RECONFIG_HARD	2

struct pinfo {
	pid_t			pid;
	char			cmd[255];
	STAILQ_ENTRY(pinfo)	entry;
};

static STAILQ_HEAD(, pinfo) childlist = STAILQ_HEAD_INITIALIZER(childlist);

static void
save_childinfo(pid_t pid, const char *path, const char **av, int ac)
{
	struct pinfo *entry;
	int i;

	entry = calloc(1, sizeof(*entry));
	if (entry == NULL)
		return;
	entry->pid = pid;
	snprintf(entry->cmd, sizeof(entry->cmd), "%s", path);
	for (i = 1; i < ac; i++)
		snprintf(entry->cmd, sizeof(entry->cmd), "%s %s",
		    entry->cmd, av[i]);
	STAILQ_INSERT_TAIL(&childlist, entry, entry);
}

static const char*
get_childinfo(pid_t pid)
{
	struct pinfo *entry;

	STAILQ_FOREACH(entry, &childlist, entry)
	    if (entry->pid == pid)
		    return (entry->cmd);
	return ("unknown");
}

static void
free_childinfo(pid_t pid)
{
	struct pinfo *entry;

	STAILQ_FOREACH(entry, &childlist, entry)
	    if (entry->pid == pid) {
		    STAILQ_REMOVE(&childlist, entry, pinfo, entry);
		    free(entry);
		    return;
	    }
}

static void
free_childlist(void)
{
	struct pinfo *entry;

	while (!STAILQ_EMPTY(&childlist)) {
		entry = STAILQ_FIRST(&childlist);
		STAILQ_REMOVE_HEAD(&childlist, entry);
		free(entry);
	}
}

static void
dummy_handler(int sig __unused)
{

}

static int
setup_signals(void)
{
	struct sigaction sigact;
	sigset_t sigmask;
	int i;

	sigemptyset(&sigmask);
	sigemptyset(&sigact.sa_mask);
	for (i = 0; siglist[i]; i++) {
		sigact.sa_flags = SA_RESTART;
		sigact.sa_handler = SIG_DFL;
		if (siglist[i] == SIGCHLD) {
			/* Need to change the handler to not be ignored */
			sigact.sa_flags |= SA_NOCLDSTOP;
			sigact.sa_handler = dummy_handler;
		}
		if (sigaction(siglist[i], &sigact, NULL) == -1) {
			report(LOG_ERR, "sigaction failed");
			return (-1);
		}
		sigaddset(&sigmask, siglist[i]);
	}
	return (sigprocmask(SIG_SETMASK, &sigmask, NULL));
}

static void
handle_sigchld(void)
{
	pid_t pid;
	int status;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (WIFEXITED(status) == 0 || WEXITSTATUS(status) != 0)
			report(LOG_INFO, "child %d exited with status %d (%s)\n",
			    pid, WIFEXITED(status) ? WEXITSTATUS(status): -1,
			    get_childinfo(pid));
		free_childinfo(pid);
	}
}

static int
check_signals(void)
{
	struct timespec sigtimeout;
	sigset_t sigmask;
	int has_int, has_hup, has_alrm;
	int signo, i;

	sigtimeout.tv_sec = 0;
	sigtimeout.tv_nsec = 0;
	sigemptyset(&sigmask);
	for (i = 0; siglist[i]; i++)
		sigaddset(&sigmask, siglist[i]);
	has_int = has_hup = 0;
	while ((signo = sigtimedwait(&sigmask, NULL, &sigtimeout)) != -1) {
		switch (signo) {
		case SIGPIPE:
			if (getpid() == mypid)
				break;
		case SIGTERM:
			cleanup_handler(signo);
			break;
		case SIGINT:
			has_int = 1;
			break;
		case SIGHUP:
			has_hup = 1;
			break;
		case SIGCHLD:
			handle_sigchld();
			break;
		case SIGALRM:
			watchdog_pending++;
		}
	}
	if (has_int || has_hup) {
		report(LOG_NOTICE, "%s reconfig requested",
		    has_int ? "hard": "soft");
		return (has_int ? RECONFIG_HARD: RECONFIG_SOFT);
	}
	return (0);
}

int
main(argc, argv, envp)
	int argc;
	char **argv, **envp;
{
	int op;
	char *cp, buf[1024];
	void *parsed_cf;
	FILE *fp;
#ifdef	HAVE_PTHREAD
	struct sched_param param;
#endif
	extern int optind, opterr;
	extern char *optarg;

	/*
	 * Set the variables to it default values.
	 */
	program_name = strippath(argv[0]);
	config_file = DEFAULT_CONFIG;
	syslog_initialized = 0;
	syslog_facility = 0;
	Cflag = 0;
	cf = NULL;
	watchdog_timeout = DEFAULT_WATCHDOG;
	watchdog_pending = 0;

	/*
	 * Go through command line switches and arguments.
	 */
	opterr = 0;
	while ((op = getopt(argc, argv, "c:CMw:vh?")) != EOF) {
		switch (op) {
		case 'c':	/* config file */
			config_file = optarg;
			break;
		case 'C':	/* parse config and exit */
			Cflag++;
			break;
		case 'M':	/* dump MIB tables */
			dump_mib_table(0); /* all tables */
			exit(0);
		case 'w':	/* watchdog timer */
			watchdog_timeout = atoi(optarg);
			break;
		case 'v':
			vers();
		case 'h':
		case '?':
		default:
			usage();
		}
	}
	cp = NULL;
	if (argv[optind] != NULL) {
		;	/* nope yet */
#ifdef	DEBUG
		cp = argv[optind];
#endif
	}

	/*
	 * Initialize configurations.
	 */
	if (*config_file != '/') {
		if (getcwd(buf, sizeof(buf)) == NULL) {
			report(LOG_ERR, "getcwd: %m");
			exit(1);
		}
		(void)strcat(buf, "/");
		(void)strcat(buf, config_file);
		config_file = strdup(buf);
	}
	if ((parsed_cf = parse_config(NULL)) == NULL)
		exit(1);
	if ((cf = (CONFIG *)malloc(sizeof(CONFIG))) == NULL) {
		report(LOG_ERR, "malloc: Out of memory");
		exit(1);
	}
	memcpy(cf, parsed_cf, sizeof(CONFIG));

#ifdef	DEBUG_MEMORY
	set_cksum();
#endif
	if (Cflag) {
		dump_config(cf);
		if (cp) printf("%s->%s\n", cp,
			       insert_variables(buf, sizeof(buf), cp, 0, NULL));
		exit(0);
	}

#ifdef	HAVE_PTHREAD
	/*
	 * Initialize threads.
	 */
	busy_threads = 0;
	main_thr = pthread_self();
	pthread_getschedparam(main_thr, &op, &param);

	pthread_attr_init(&joinable_attr);
	pthread_attr_setdetachstate(&joinable_attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setschedparam(&joinable_attr, &param);

	pthread_attr_init(&detached_attr);
	pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);
	param.sched_priority /= 2; /* lower priority for detached childs */
	pthread_attr_setschedparam(&detached_attr, &param);
#endif

	if (rootdir_configure(cf) < 0)
		exit(1);

#ifndef	DEBUG
	/*
	 * Start as daemon.
	 */
	daemon(1, 0);
	if (syslog_initialized == 0) {
		if (syslog_facility == 0)
			syslog_facility = LOG_DAEMON;
		openlog(program_name, LOG_PID|LOG_CONS, syslog_facility);
		syslog_initialized = 1;
	}
#endif /* !DEBUG */
	mypid = getpid();

	if (trap_init(cf->enable_traps > 0) < 0)
		exit(1);
	if (netstate_init(cf->ns_port) < 0)
		exit(1);

	/*
	 * Make pid file.
	 */
	snprintf(buf, sizeof(buf), "%s/%s.pid", PIDFILE_PATH, program_name);
	if ((fp = fopen(buf, "w")) != NULL) {
		fprintf(fp, "%d\n", (int)mypid);
		fclose(fp);
	}

#ifndef	HAVE_PTHREAD
	/*
	 * Save start and extend of argv.
	 */
	Argv = argv;
	while (*envp) envp++;
	LastArgv = envp[-1] + strlen(envp[-1]);
#endif

	report(LOG_NOTICE, "started");

	/*
	 * Prepare signal mask and handlers.
	 */
	if (setup_signals() != 0)
		exit(1);

	if (watchdog_timeout)
		set_timer(watchdog_timeout, NULL);
	start_object_list(cf->target);
	mainloop();

	/* NOT REACHED */
	cleanup_handler(0);
}

void
watchdog(passed)
	int passed; /* seconds since previous call */
{
	OBJECT *target, *service;
	TIMEVAL now, tv;

	dprintf(("watchdog(%d) called\n", passed));

	watchdog_pending = 0;
	if (!passed)
		return;

	gettimeofday(&now, NULL);
	for (target = cf->target; target; target = target->next) {
		if (target->polling > 0 && target->method_list) {
			tv = now;
			tv_sub(&tv, &target->last_request);
			if (tv.tv_sec >= passed) {
				int lost = !remove_event(NULL, target);

				target->state = 0; /* set NONE state */
				reset_statistic(target->var_list, 1);

				object_stop(target);
				for (service = target->service; service; service = service->next)
					object_stop(service);

				set_none_state(target);
				save_object_state(target);

				report(LOG_WARNING, "watchdog %s polling at \"%s\"",
				       lost ? "resume" : "restart", target->name);

				object_init(target);
				for (service = target->service; service; service = service->next) {
					service->ip_addr = target->ip_addr;
					service->parent = target;
					object_init(service);
				}
				start_method_list(target);
			}
		} else if (trap_sock < 0) {
			report(LOG_WARNING, "No polling configured at \"%s\"?",
			       target->name);
		}
	}
}

void
mainloop(void)
{
	struct sockaddr_in from;
	fd_set readfds, writefds;
	int op, nfds, save_errno, ret;
	TIMEVAL timeout;
	void *parsed_cf;
	TRAP *trap;
#ifdef	HAVE_PTHREAD
	pthread_t reconfig_thr, netstate_thr;
#endif
	for (;;) {
		if (cf->flags & CONFIG_F_SOFTRECONFIG)
			finish_soft_reconfig();
		ret = check_signals();
		/* Reconfig requested? */
		if (ret == RECONFIG_SOFT &&
		    (cf->flags & CONFIG_F_SOFTRECONFIG) != 0) {
			report(LOG_NOTICE, "soft reconfig already in progress");
		} else if (ret != 0) {
			parsed_cf = parse_config(NULL);
			if (parsed_cf == NULL)
				report(LOG_NOTICE,
				    "failed to parse config file");
			else if (ret == RECONFIG_HARD)
				hard_reconfig((CONFIG *)parsed_cf);
			else
				soft_reconfig((CONFIG *)parsed_cf);
		}
		if (session_busy(NULL) == NULL && watchdog_pending != 0)
			watchdog(watchdog_pending * watchdog_timeout);
		nfds = 0;
		FD_ZERO(&readfds);
		if (trap_sock != -1) {
			nfds = trap_sock + 1;
			FD_SET(trap_sock, &readfds);
		}
		if (netstate_sock != -1) {
			nfds = netstate_sock + 1;
			FD_SET(netstate_sock, &readfds);
		}
		FD_ZERO(&writefds);
		op = select_event(&timeout);
		session_select(&nfds, &readfds, &writefds, &timeout, &op);
		op = select(nfds, &readfds, &writefds, 0, op ? 0 : &timeout);
		if (op > 0) {
			trace(("\t select operate\n"));

			session_operate(&readfds, &writefds);
			if (trap_sock != -1 &&
			    FD_ISSET(trap_sock, &readfds)) {
				if ((trap = snmp_trap(trap_sock, cf->source_traps > 0)) != NULL) {
					trap_parse(trap);
					trap_free(trap);
				}
			}
			if (netstate_sock != -1 &&
			    FD_ISSET(netstate_sock, &readfds)) {
				socklen_t from_len = sizeof(from);
				if ((op = accept(netstate_sock, (struct sockaddr *)&from, &from_len)) != -1) {
#ifdef	HAVE_PTHREAD
					pthread_create(&netstate_thr, &detached_attr, netstate_serve, (void *)op);
#else
					if (fork() == 0) { /* child */
						close(netstate_sock);
						set_timer(0, NULL);
						netstate_serve((void *)(intptr_t)op);
						_exit(0);
					}
					close(op); /* parent */
#endif
				} else	report(LOG_ERR, "accept: %m");
			}
		} else if (op == 0) {
			trace(("\t select timedout\n"));

			session_timeout();
		} else {
			trace(("\t select interrupted\n"));
		}
	}
	/* NOT REACHED */
}

int
set_timer(int timeout, void (*func)(int))
{
	struct sigaction sigact;
	struct itimerval it, rest;

	if (func != NULL) {
		sigaction(SIGALRM, NULL, &sigact);
		sigact.sa_handler = func;
		sigaction(SIGALRM, &sigact, NULL);
	}
	it.it_value.tv_sec = it.it_interval.tv_sec = timeout;
	it.it_value.tv_usec = it.it_interval.tv_usec = 0;
	setitimer(ITIMER_REAL, &it, &rest);
	return (rest.it_value.tv_sec);
}

void
object_init(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (!method->init) {
			errno = EINVAL;
			report(LOG_ERR, "object_init(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
		if ((*method->init)(object, method) < 0) {
			report(LOG_ERR, "object_init(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
	}
}

void
object_stop(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (!method->stop) {
			errno = EINVAL;
			report(LOG_ERR, "object_stop(%s/%s): %m",
			       object->name, method->name);
			cleanup_handler(0);
		}
		(*method->stop)(method);
	}
}

int
object_busy(object)
	OBJECT *object;
{
	METHOD *method;

	for (method = object->method_list; method; method = method->next) {
		if (method->sd && session_busy(method->sd))
			return 1;
	}
	return 0;
}

void
start_method_list(arg)
	void *arg;
{
	OBJECT *object = (OBJECT *)arg;
	METHOD *method;
	TIMEVAL tv;
	int busy;

	/* sanity check */
	if (!object) {
		errno = EINVAL;
		report(LOG_ERR, "start_method_list(%s): %m", object->name);
		cleanup_handler(0);
	}

	method = object->method_list;
	if (!method) {
		report(LOG_NOTICE, "No polling methods at \"%s\"", object->name);
		/* save_object_data(object); -- no periodic saving for traps */
		return;
	}

	/* current timestamp */
	gettimeofday(&tv, NULL);

	busy = object_busy(object);
	if (busy || (!object->parent && !check_netpath(object))) {
		if (busy) {
			if (object->parent)
				report(LOG_DEBUG, "Polling methods at \"%s/%s\" still busy -- aborted!",
				       object->parent->name, object->name);
			else	report(LOG_DEBUG, "Polling methods at \"%s\" still busy -- aborted!",
				       object->name);
			object_stop(object);
		}

		/* set the NONE state on the object */
		object->state = 0;
		reset_statistic(object->var_list, 1);

		/* set the NONE state to all subobjects */
		set_none_state(object);

		/* save state changes if required */
		save_object_state(object);

		/* stop data saving on the object */
		remove_event(save_object_data, object);

		/* keep normal polling */
		if (object->polling > 0) {
#ifdef DEBUG
			static char buf[50], buf2[50];
			strftime(buf, sizeof(buf), "%T", localtime(&tv.tv_sec));
			snprintf(buf, sizeof(buf), "%s.%03d", buf, tv.tv_usec / 1000);
#endif
			tv_sharp(&tv, object->polling, object_offset(object, object->polling));
#ifdef DEBUG
			strftime(buf2, sizeof(buf2), "%T", localtime(&tv.tv_sec));
			dprintf(("start_method_list(%s [%p]): sheduled at %s (was %s.%03d)\n",
			    object->name, arg, buf2, buf, tv.tv_usec / 1000));
#endif
			add_event(&tv, start_method_list, object);
		}
		return;
	}

	dprintf(("start_method_list(%s)\n", object->name));

	object->prev_request = object->last_request;
	object->last_request = tv; /* last request timestamp */

	(*method->start)(method);
}

void
method_finished(object, method, diag, ok)
	OBJECT *object;
	METHOD *method;
	char *diag;
	int ok;
{
	TIMEVAL tv;
	VARIABLE *var;
	OBJECT *service;
	INTERFACE *interface;
	BGP_AS	*bgp;
	BGP_PEER *bgp_peer;
	ENV_MON	*env;
	ENV_GAUGE *gauge;
	char *buf;
	ssize_t buf_len; 

	/* current timestamp */
	gettimeofday(&tv, NULL);
	/* save method diagnostic */
	if (method && diag &&
	    (var = get_var(object->var_list, method->name)) != NULL) {
		str2var(var, diag);
		/* save method return value */
		buf_len = strlen(method->name) + 4;
		buf = malloc(buf_len);
		if (buf != NULL) {
			snprintf(buf, buf_len, "%s.ok", method->name);
			if ((var = get_var(object->var_list, buf)) != NULL)
				set_var(var, INTEGER, &ok, sizeof(ok));
			free(buf);
		} else {
			report(LOG_ERR, "method_finished(%s!):.ok malloc: %m",
			    object->name, method->name);
		}
	}
	if (compatibility_flag != 0) {
		if (!method || !ok) {
			/* update object operational status */
			object->prev_state = object->state;
			object->state = STATE_DOWN;
			if (object->state != object->prev_state)
				object->last_change = tv;
#ifdef	DEBUG
			if (object->prev_state != STATE_DOWN)
				dprintf(("object \"%s\" change state to DOWN\n",
				    object->name));
#endif
			/* stop anything here */
			object_stop(object);
			for (service = object->service; service;
			    service = service->next)
				object_stop(service);

			set_none_state(object);
			save_object_state(object);
			/* stop data saving on the object */
			remove_event(save_object_data, object);
			/* keep touching the object if required */
			if (object->polling > 0) {
#ifdef DEBUG
				static char buf[50], buf2[50];
				strftime(buf, sizeof(buf), "%T",
				    localtime(&tv.tv_sec));
				snprintf(buf, sizeof(buf), "%s.%03d", buf,
				    tv.tv_usec / 1000);
#endif
				tv.tv_sec += object->polling / 2;
#ifdef DEBUG
				strftime(buf2, sizeof(buf2), "%T",
				    localtime(&tv.tv_sec));
				dprintf(("method_finished(%s [%p])[0]: "
				    "sheduled at %s (was %s.%03d)\n",
				    object->name, object, buf2, buf,
				    tv.tv_usec / 1000));
#endif
				add_event(&tv, start_method_list, object);
			}
			return;
		}

		if (method->next) {
			method = method->next;
			(*method->start)(method);
			return;
		}

		object->prev_reply = object->last_reply;
		object->last_reply = tv; /* last reply timestamp */

		/* update object operational status */
		object->prev_state = object->state;
		object->state = STATE_UP;
	} else {
		if (!method || !ok) {
			object->mths_fail++;
		}  else {
			object->mths_ok++;
		}

		if (method && method->next) {
			/*  Advance to next object method.  */
			method = method->next;
			(*method->start)(method);
			return;
		}
		/* Method list done. */
		object->prev_reply = object->last_reply;
		object->last_reply = tv; /* last reply timestamp */

		/* update object operational status */
		object->prev_state = object->state;
		if (object->mths_ok == 0) {
			object->state = STATE_DOWN;
			if (object->state != object->prev_state)
				object->last_change = tv;
#ifdef	DEBUG
			if (object->prev_state != STATE_DOWN)
				dprintf(("object \"%s\" change state to DOWN\n",
				    object->name));
#endif
			/* stop anything here */
			object_stop(object);

			for (service = object->service; service;
			    service = service->next)
				object_stop(service);

			set_none_state(object);

			save_object_state(object);

			/* stop data saving on the object */
			remove_event(save_object_data, object);

			/* keep touching the object if required */
			if (object->polling > 0) {
#ifdef DEBUG
				static char buf[50], buf2[50];
				strftime(buf, sizeof(buf), "%T",
				    localtime(&tv.tv_sec));
				snprintf(buf, sizeof(buf), "%s.%03d", buf,
				    tv.tv_usec / 1000);
#endif
				tv.tv_sec += object->polling / 2;
#ifdef DEBUG
				strftime(buf2, sizeof(buf2), "%T",
				    localtime(&tv.tv_sec));
				dprintf(("method_finished(%s [%p])[1]: "
				    "sheduled at %s (was %s.%03d)\n",
				    object->name, object, buf2, buf,
				    tv.tv_usec / 1000));
#endif
				add_event(&tv, start_method_list, object);
			}
			return;
		} else {
			if ( object->mths_fail ) {
				object->state = STATE_DEGRADED;
			} else {
				object->state = STATE_UP;
				for (service = object->service; service;
				    service = service->next) {
					if ((service->flags &
					    OBJECT_F_IGNORE_STATE) != 0 &&
					    service->state != STATE_UP) {
						object->state = STATE_WARNING;
						break;
					}
				}
				for (interface = object->interface; interface;
				    interface = interface->next) {
					if ((interface->flags &
					    INTERFACE_F_IGNORE_STATE) != 0 &&
					    interface->state != STATE_UP) {
						object->state = STATE_WARNING;
						break;
					}
				}
				for (bgp = object->bgp; bgp; bgp = bgp->next) {
					if (bgp->flags & BGPAS_F_IGNORE_STATE)
						continue;
					for (bgp_peer = bgp->peer; bgp_peer;
					    bgp_peer = bgp_peer->next) {
						if (bgp_peer->state !=
						    BGP_ESTABLISHED) {
							object->state =
							    STATE_WARNING;
							break;
						}
					}
					if (object->state == STATE_WARNING)
						break;
				}
				for (env = object->env; env; env = env->next) {
					if (env->flags & ENVMON_F_IGNORE_STATE)
						continue;
					for(gauge = env->gauge; gauge;
					    gauge=gauge->next) {
						if (gauge->state !=
						    ENV_NORMAL &&
						    gauge->state !=
						    ENV_NOTPRESENT) {
							object->state =
							    STATE_WARNING;
							break;
						}
					}
					if (object->state == STATE_WARNING)
						break;
				}
			}
		}

		object->mths_ok = 0;
		object->mths_fail = 0;
		object->smths_fail = 0;
	}
	if (object->state != object->prev_state)
		object->last_change = tv;

#ifdef	DEBUG
	if (object->prev_state != STATE_UP)
		dprintf(("object \"%s\" change state to UP\n", object->name));
#endif

	save_object_state(object);

	/* to avoid network flooding at startup */
	if (timerisset(&object->prev_request)) {
		/* if list of subobject present start it here */
		for (service = object->service; service;
		    service = service->next)
			start_method_list(service);
	}

	if (object->polling > 0) {
		/* start data saving if required */
		if (object->saving > 0 &&
		    !find_event(save_object_data, object)) {
			if (!object->sync) {
				save_object_data(object);
			} else if (--object->sync == 0) {
				object->sync = object->saving / object->polling;
				save_object_data(object);
			}
		}
#ifdef DEBUG
		static char buf[50], buf2[50];
		strftime(buf, sizeof(buf), "%T", localtime(&tv.tv_sec));
		snprintf(buf, sizeof(buf), "%s.%03d", buf, tv.tv_usec / 1000);
#endif
		/* keep polling the object */
		tv_sharp(&tv, object->polling,
		    object_offset(object, object->polling));
#ifdef DEBUG
		strftime(buf2, sizeof(buf2), "%T", localtime(&tv.tv_sec));
		dprintf(("method_finished(%s [%p])[2]: "
		    "sheduled at %s (was %s.%03d)\n", object->name, object,
		    buf2, buf, tv.tv_usec / 1000));
#endif
		add_event(&tv, start_method_list, object);
	}
}

void
start_object_list(object)
	OBJECT *object;
{
	OBJECT *target, *service;

	/* initialize sessions for configured objects */
	for (target = object; target; target = target->next) {
		object_init(target);

		/* and it subobjects */
		for (service = target->service; service; service = service->next) {
			service->ip_addr = target->ip_addr;
			service->parent = target;

			object_init(service);
		}
	}

	/* start sessions for configured objects */
	for (target = object; target; target = target->next)
		start_method_list(target);
}

#ifndef	HAVE_PTHREAD
/*
 * clobber argv so ps will show what we're doing.
 * stolen from sendmail
 */
void
proctitle(str)
	char *str;
{
	int i;
	char *cp, buf[1024];

	cp = Argv[0];
	sprintf(buf, "%s: %s", program_name, str);
	i = strlen(buf);
	if (i > LastArgv - cp - 1)
		i = LastArgv - cp - 1;
	(void)strncpy(cp, buf, i);
	cp[i] = '\0';
	cp += i;
	while (cp < LastArgv) *cp++ = ' ';
}
#endif

#define	MAX_ARGS	50

void
execfile(ispipe, file, args)
	int ispipe;
	char *file, *args;
{
	sigset_t sigmask;
	int ac = 0, fildes[2];
	char *av[MAX_ARGS+2];
	extern char **environ;
	pid_t child;

	dprintf(("execfile: %s %s\n", file, args ? args : "null"));

	if (ispipe && pipe(fildes) < 0) {
		report(LOG_ERR, "pipe: %m");
		return;
	}

	if (*file != '/') {
		av[ac++] = "sh";
		av[ac++] = "-c";
		av[ac++] = file;
#ifdef	_PATH_BSHELL
		file = _PATH_BSHELL;
#else
		file = "/bin/sh";
#endif
	} else	av[ac++] = strippath(file);
	if (!ispipe)
		ac += make_argv(args, (char ***) &av[ac], MAX_ARGS);
	av[ac] = NULL;

	child = vfork();
	switch (child) {
	case -1:
		report(LOG_ERR, "fork: %m");
		if (ispipe) {
			close(fildes[0]);
			close(fildes[1]);
		}
		return;
	case 0:		/* child */
		if (ispipe) {
			if (fildes[0] != STDIN_FILENO) {
				dup2(fildes[0], STDIN_FILENO);
				close(fildes[0]);
			}
			close(fildes[1]);
		}

		/* child would be terminated by signals */
		sigprocmask(SIG_SETMASK, NULL, &sigmask);
		sigprocmask(SIG_UNBLOCK, &sigmask, NULL);
		close(netstate_sock);
		
		/* make session leader to be able killpg() latter */
		setsid();

		if ( cf->chrootdir) {
		    if ( chroot( cf->chrootdir ) < 0 ) {
			report(LOG_ERR, "chroot %s: %s", cf->chrootdir,strerror(*(__error())) );
			_exit(127);
		    }
		}
		if ( setgid(cf->gid) < 0 ) {
		    report(LOG_ERR, "setgid %s[%d]: %s", cf->groupname, cf->gid, strerror(*(__error())) );
		    _exit(127);
		}			
		if ( (cf->uid != 0) & (setuid(cf->uid) < 0) ) {
		    report(LOG_ERR, "setuid %s[%d]: %s", cf->username, cf->uid, strerror(*(__error())) );
		    _exit(127);
		}		
		execve(file, av, environ);
		report(LOG_ERR, "execve %s: %m", file);
		_exit(127);
	}
	/* parent */
	save_childinfo(child, file, (const char **)av, ac);

	if (ispipe) {
		FILE *fp;

		close(fildes[0]);
		if ((fp = fdopen(fildes[1], "w")) != NULL) {
			fprintf(fp, "%s\n", args);
			fclose(fp);
		} else	report(LOG_ERR, "fdopen: %m");
	}
}

static void
cleanup_handler(sig)
	int sig;
{
#ifdef	HAVE_PTHREAD
	if (pthread_self() == main_thr)
#else
	if (getpid() == mypid)
#endif
	{
		char pidfile[100];
		snprintf(pidfile, sizeof(pidfile), "%s/%s.pid", PIDFILE_PATH, program_name);
		(void)unlink(pidfile);
		report(LOG_CRIT, "aborted by signal %d", sig);
		free_childlist();
	} else	report(LOG_INFO, "interrupted by signal %d", sig);
	exit(1);
}

#include <zlib.h>

void
vers()
{
	fprintf(stderr, "\n%s Version %s\n\tcompiled for %s with\n %s\n",
		package, version, target, compiled);
#ifdef	HAVE_PTHREAD
	fprintf(stderr, "\tusing POSIX threads\n");
#endif
#ifdef	HAVE_DMALLOC_H
	fprintf(stderr, "\twith dmalloc library\n");
#endif
	fprintf(stderr, "\twith zlib version %s\n", ZLIB_VERSION);
	fprintf(stderr, "\n%s\n", copyright);
	fprintf(stderr, "%s support list: netmon@service.risp.ru\n\
\tsend `subscribe netmon' to majordomo@service.risp.ru\n", package);
	exit(0);
}

void
usage()
{
	fprintf(stderr, "usage:\n\
  %s [-c file] [-C] [-M] [-v] [-w]\n", program_name);

	fprintf(stderr, "where:\n\
  -c file  Config file, default %s\n\
  -C       Check config file for syntax errors\n\
  -M       Dump known SNMP MIB tables and exit\n\
  -v       Print version number, compile-time options, and exit\n\
  -w sec   Watch dog timer (0=disable), default %d sec.\n",
		DEFAULT_CONFIG, DEFAULT_WATCHDOG);

	exit(1);
}
